# creating the main executor!
global_controller = None

def set_main_controller(cr):
    global global_controller
    global_controller = cr

def get_main_controller():
    return global_controller





"""
    The Runner interface is as follows:
    The runner is in charge of syncing files input/output, meaning if it is running remotely, it needs to send the input and bring back the output.

     - available_slots() returns a positive number with available resources, returns 0 if at capacity
     - add(task): this starts the job right away, if we are at capacity we should throw an error, the controller should be checking capacity
     - loop: a loop that runs in the event-loop, when a task completes, it calls the controller `add_completed`.

"""
from abc import ABC, abstractmethod

import asyncio
import logging
import queue
import subprocess
from datetime import datetime
from omegaconf import OmegaConf
import os, tempfile
import time

class AbstractRunner(ABC):

    @abstractmethod
    def size():
        pass

    @abstractmethod
    def available_slots():
        pass

    @abstractmethod
    def add(self, task):
        pass


class CommandRunner(AbstractRunner):

    def __init__(self, conf):
        conf = OmegaConf.create(conf)
        self.max_proc = conf.get('maxsize',4)
        self.processes = {}
        self.logger = logging.getLogger(self.__class__.__name__)

    def size(self):
        return(len(self.processes))

    def available_slots(self):
        return self.max_proc - len(self.processes)

    """
        Start tasks
    """
    def add(self, task):

        if task.quiet:
            subp = subprocess.Popen(task.get_command(),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT)
            # @fixme, log to file here instead of nowhere
        else:
            subp = subprocess.Popen(task.get_command())
                #stdout=subprocess.DEVNULL,
                #stderr=subprocess.STDOUT)

        self.processes[task.hash] = {
            "proc" : subp, 
            "task": task , 
            'start_time': datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        }                

    """
        Continuously checks on the tasks
    """
    async def loop(self, controller):
        while True:
            self.update(controller)                
            await asyncio.sleep(0.1)

    def update(self,controller):
        to_remove = []
        for (k,p) in self.processes.items():
            poll_val = p["proc"].poll()
            if poll_val is not None:
                to_remove.append(k)

        for k in to_remove:
            task = self.processes[k]["task"]
            del self.processes[k]       
            controller.add_completed( task )


class HpcRunner(AbstractRunner):

    script_template = """
    #PBS -N {name}
    #PBS -j oe
    #PBS -V
    #PBS -l procs={procs},mem={mem}
    #PBS -l walltime=12:00:00

    module load julia/1.6.1
    cd {wd}

    {cmd}
    """

    def __init__(self, conf):
        conf = OmegaConf.create(conf)
        self.max_proc = conf.maxsize
        self.processes = {}
        self.job_params = {'procs':1, 'mem' : "16Gb", 'name':'psub'}

    def size(self):
        return(len(self.processes))

    def available_slots(self):
        return self.max_proc - len(self.processes)

    def add(self, task):

        # create the script
        script_content = self.script_template.format(
            name = "sf-{}".format(task.uid),
            mem = task.mem,
            procs = task.ncore,
            wd = os.getcwd(), 
            cmd = " ".join(task.get_command()) )

        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp_script_filename = tmp.name
        tmp.write(script_content.encode())
        tmp.close()

        command = ["qsub", tmp_script_filename]
        try:
            output = subprocess.check_output(command).decode()
            JOB_ID = output.replace("\n","")
            
            self.processes[task.hash] = {"JOB_ID" : JOB_ID, "job": task, "status":"S", 'start':time.time()}          
        except subprocess.CalledProcessError as e:
            self.processes[task.hash] = {"JOB_ID" : JOB_ID, "job": task, "status":"F", 'start':time.time()}          

    def update(self, controller):

        # checking job status
        output = subprocess.check_output("qstat").decode()
        lines = output.split("\n") 
        job_status = {}
        for l in lines[2:]:                    
            vals = l.split()
            if len(vals)<3:
                continue
            job_status[vals[0]] = {'status':vals[4]}

        to_remove = []
        for (k,p) in self.processes.items():

            if p["JOB_ID"] not in job_status.keys():

                # we expire the job if longer than 5 minutes
                if time.time() - p["start"] > 5*60:
                    job_status[p["JOB_ID"]] = {'status':'C'}
                else:
                    continue

            if job_status[p["JOB_ID"]]['status'] == 'C':
                to_remove.append(k)

        for k in to_remove:
            del self.processes[k]
            controller.add_completed( p["job"] )

    async def loop(self,controller):
        while True:
            self.update(controller)                 
            await asyncio.sleep(2)

EXECUTORS = {    
    "local" : CommandRunner,
    "hpc" : HpcRunner,
}




import hashlib
import asyncio
import shlex

from .glob import get_main_controller


"""

note: 
 - use input and output for dependencies and generated stuff
 - use set_args and set_return for input parameters and return values that should be collected

"""
class Task:
    cmd =""
    uid=""
    mem="32Gb"
    ncore="10"
    retry=0

    """
    Allows to construct the class with several options!
    """
    def __init__(self, **kwargs):

        self.fut = None
        self.state = "init"

        if "cmd" in kwargs.keys():
            # we check if we have a string, in which case we try to split it
            if isinstance(kwargs["cmd"], str):
                self.cmd = shlex.split(kwargs["cmd"])
            else:
                self.cmd  = kwargs["cmd"]

        if "controller" in kwargs.keys():
            self.controller  = kwargs["controller"]
        else:
            self.controller = get_main_controller()

        if "outputs" in kwargs.keys():
            self.output_file  = kwargs["outputs"]
        else:
            self.output_file = ""

        if "inputs" in kwargs.keys():
            self.deps  = kwargs["inputs"]
            if isinstance(kwargs["inputs"], str):
                self.deps = [kwargs["inputs"]]
            else:
                self.deps  = kwargs["inputs"]
        else:
            self.deps=[]

        if "name" in kwargs.keys():
            self.uid  = kwargs["name"]
        else:
            self.uid = ""

        if "quiet" in kwargs.keys():
            self.quiet  = kwargs["quiet"]
        else:
            self.quiet = True

        self.hash = ""
        self.return_file = ""
        self.props = {}

    def __await__(self):
        # we need to check if the task has been scheduled
        if not self.is_scheduled():
            self.schedule()

        return self.fut.__await__()

    def schedule(self):
        self.fut = asyncio.get_event_loop().create_future()
        self.state = "scheduled"
        self.hash = hashlib.md5("".join(self.get_command()).encode()).hexdigest()
        self.controller.add(self)
        return(self)

    def result(self, return_file):
        self.return_file = return_file    
        return self  

    def output(self, output_file):
        self.output_file = output_file    
        return self     

    def retry(self, retry):
        self.retry = retry    
        return self     

    def input(self, input_file):
        # check that the input file actually exist and extract its values, too soon?
        self.input_file = input_file  
        return self  

    def add_deps(self, deps):

        if isinstance(deps, str):
            self.deps.append(deps)
            return(self)

        if not hasattr(deps, "__iter__"): #returns True if type of iterable - same problem with strings
            deps = list(deps)
        
        for dep in deps:
            self.deps.append(dep)

        return(self)

    def name(self,name):
        self.name = name

    def uid(self,uid):
        self.uid = uid
        return self

    def get_command(self):
        return self.cmd

    def set_prop(self,name,value):
        self.props[name]=value

    def get_prop(self,name):
        return(self.props[name])

    def get_outputs(self):
        if self.output_file=="":
            return([])

        return([self.output_file])

    def set_state_scheduled(self):
        self.state = "scheduled"

    def is_scheduled(self):
        return(self.state == "scheduled")

    def set_state_completed(self):
        self.state = "completed"

    def set_completed(self):
        self.fut.set_result(self)



import click
import asyncio

# loading the use file
import importlib.util
from inspect import getmembers, isfunction

import scriptflow.core as sf
import os
import logging

from .glob import get_main_controller

# loading the local sflow.py file
def load_local_file(filename):
    spec = importlib.util.spec_from_file_location("", os.getcwd() + "/" + filename)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    return foo

def _handle_task_result(task: asyncio.Task) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    except Exception:  # pylint: disable=broad-except
        logging.exception('Exception raised by task = %r', task)

"""
    Main
"""
async def main(func, controller:sf.Controller):

    task_controller = asyncio.create_task( controller.start_loops() )
    task_controller.add_done_callback(_handle_task_result)

    os.makedirs('.sf', exist_ok=True)

    await func()      
    
    # let the other guys finish
    await asyncio.sleep(1)
    controller.last_word()


@click.group()
def cli():
    pass

@cli.command()
@click.argument('name')
@click.option('-f', '--force', 'force')
def run(name="",force=None,load="sflow.py"):

    foo = load_local_file("sflow.py")

    func_names = {k:v for (k,v) in getmembers(foo, isfunction)}
    flows = [w.replace("flow_","") for w in func_names.keys() if w.startswith("flow_")]

    if name not in flows:
        print("Flow '{}' is not available, choose from: {}".format(name, ", ".join(flows)))
        return()

    func = func_names["flow_{}".format(name)]

    controller = get_main_controller()
    controller.set_force_string(force)
    
    asyncio.run(main(func, controller))

# if __name__ == '__main__':
#     cli()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scriptflow']

package_data = \
{'': ['*']}

install_requires = \
['asyncssh>=2.9.0,<3.0.0',
 'click>=8.0.3,<9.0.0',
 'omegaconf>=2.1.1,<3.0.0',
 'pytest>=7.1.1,<8.0.0',
 'requests>=2.27.1,<3.0.0',
 'rich>=11.0.0,<12.0.0',
 'tinydb>=4.7.0,<5.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['scriptflow = scriptflow.cli:cli']}

setup_kwargs = {
    'name': 'scriptflow',
    'version': '0.2.4',
    'description': 'Like a makefile but in python, a stripped-down system of Airflow or Luigi',
    'long_description': '# scriptflow\n\n[![CircleCI](https://circleci.com/gh/tlamadon/scriptflow/tree/main.svg?style=svg)](https://circleci.com/gh/tlamadon/scriptflow/tree/main) [![PyPI version](https://badge.fury.io/py/scriptflow.svg)](https://badge.fury.io/py/scriptflow) [![codecov](https://codecov.io/gh/tlamadon/scriptflow/branch/main/graph/badge.svg?token=0E8J7635HD)](https://codecov.io/gh/tlamadon/scriptflow)\n\nSmall library that allows scheduling scripts asyncrhonously on different platforms. Think of it as a Make when you can write the dependencies as python code, and that can run locally, on an HPC or in the cloud (cloud is not implemented just yet).\n\nThe status is very experimental. I will likely be changing the interface as I go. \n\n## Goals:\n\n - [x] works on windows / osx / linux\n - [x] describe dependencies as python code (using await/async)\n - [x] describe scripts with input/output as code\n - [x] clean terminal feedback (using rich)\n - [x] task retry\n - [x] check that output was generated \n - [x] notifications (using light webserver at [scriptflow.lamadon.com](http://scriptflow.lamadon.com/) )\n - [x] send status to central web service\n - [x] resume flows\n - [ ] clean output\n - [ ] named runs\n - [x] store run information\n - [x] output diagnostic / reporting (tracing how files were created)\n - [x] simpler interface with default head executor and awaitable tasks\n - [x] skip computation based on timestamp of inputs and outpus\n - [ ] load and store tasks results\n - [ ] remove existing output of task if task is started (issue with failing tasks that look like they worked)\n - executors :\n   - [x] local excutor using subprocess \n   - [x] HPC excutor (monitoring qsub) \n   - [ ] docker Executor \n   - [ ] aws executor (probably using Ray)\n   - [ ] dask executor  \n - [x] add check on qsub return values\n - [x] select flow by name from terminal \n - [ ] ? scripts can create tasks, not sure how to await them. \n - reporting:\n   - [ ] input and output hashes\n   - [x] start and end datetimes\n - notification system\n   - [x] allow to send messages\n   - [ ] allow for runs\n   - [ ] allow to send messages with html content like images\n - writing tasks and flows \n   - [ ] cache flows in addition to caching tasks (avoid same task getting scheduled from 2 places)\n   - [ ] a functional api for task creation with hooks\n   - [ ] a functional api for flows\n   - [ ] controller could parse the log file for results (looking for specific triggers)\n   - [ ] allow for glob output/input\n   - [ ] provide simple toml/json interface for simple tasks and flows\n   - [x] use `shlex` to parse command from strings\n - cli\n   - [ ] pass arguments to flows \n   - [ ] create portable executable\n\n\n## Simple flow example:\n\nCreate a file `sflow.py` with:\n\n```python\nimport scriptflow as sf\n\n# set main options\nsf.init({\n    "executors":{\n        "local": {\n            "maxsize" : 5\n        } \n    },\n    \'debug\':True\n})\n\n# example of a simple step that combines outcomes\ndef step2_combine_file():\n    with open(\'test_1.txt\') as f:\n        a = int(f.readlines()[0])\n    with open(\'test_2.txt\') as f:\n        b = int(f.readlines()[0])\n    with open(\'final.txt\',\'w\') as f:\n        f.write("{}\\n".format(a+b))\n\n# define a flow called sleepit\nasync def flow_sleepit():\n\n    i=1\n    t1 = sf.Task(\n      cmd    = f"""python -c "import time; time.sleep(5); open(\'test_{i}.txt\',\'w\').write(\'5\');" """,\n      output = f"test_{i}.txt",\n      name   = f"solve-{i}")\n\n    i=2\n    t1 = sf.Task(\n      cmd    = f"""python -c "import time; time.sleep(5); open(\'test_{i}.txt\',\'w\').write(\'5\');" """,\n      output = f"test_{i}.txt",\n      name   = f"solve-{i}")\n\n    await sf.bag(t1,t2)\n\n    tfinal = sf.Task("python -c \'import sflow; sflow.step2_combine_file()\'")\n    tfinal.output(f"final.txt").uid(f"final").add_deps([t1.output_file,t2.output_file])\n    await tfinal\n```        \n\nthen create a local env, activate, install and run!\n\n```shell\npython3 -m venv env\nsource env/bin/activate\npip install scriptflow\nscritpflow run sleepit\n```\n\n## Life cycle of a task\n\n1. the task object is created. All properties can be edited.\n2. the task is sent to an executor. At this point, the properties of the task are frozen. They can be read, copied but not changed. A unique ID id created from the task from its command and its inputs. The task can be sent by using the `start()` method, or it will be sent automatically when awaited.\n3. the task is awaited, and hence execution is blocked until the task is finished. Nothing can be done at that stage. Again, the task is automatically sent at this stage if it has not be done before. Also note that several tasks can be awaited in parallel by bagging them with `sf.bag(...)`.\n4. the task is completed, the await returns. The task has now it\'s output attached to it, it can be used in the creation of other tasks.\n\n## Inspiration / Alternatives\n\nI have tried to use the following three alternatives which are all truly excelent!\n\n - [pydoit](https://pydoit.org/)\n - [nextflow](https://www.nextflow.io/)\n - [snakemake](https://snakemake.readthedocs.io/en/stable/)\n\nThere were use cases that I could not implement cleanly in the dataflow model of nextflow. I didn\'t like that snakemake relied on file names to trigger rules, I was constently juggling complicated file names. Pydoit is really great, but I couldn\'t find how to extend it to build my own executor, and I always found myself confused writing new tasks and dealing with dependencies. \n\n## Developing\n\nthe package is managed using poetry, install poetry first then \n\n```\npoetry install\n\n# run example\ncd examples/simple-local\npoetry run scriptflow run sleepit\n\n# run tests with coverate\npoetry run python -m pytest --cov=scriptflow\npoetry run coverage xml\npoetry run codecov -t <token>\n\n```\n\n\n\n\n### Docker images to try the different schedulers\n\n - [PBS](https://openpbs.atlassian.net/wiki/spaces/PBSPro/pages/79298561/Using+Docker+to+Instantiate+PBS)\n - [slurm](https://medium.com/analytics-vidhya/slurm-cluster-with-docker-9f242deee601)\n=======\n',
    'author': 'Thibaut Lamadon',
    'author_email': 'thibaut.lamadon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tlamadon/scriptflow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# simple script flow

import subprocess
import os
import time
#import toml
import asyncio
import tempfile
from datetime import datetime
import asyncio
import logging
import queue
import requests
import shlex
import fnmatch

from rich.console import Console
from time import sleep
from os.path import exists
from tinydb import TinyDB, Query
from omegaconf import OmegaConf

from .task import Task
from .runners import *
from .glob import *

console = Console()

# print(toml.load("jmp.toml"))

# with open('jmp.toml', 'w') as f:
#     new_toml_string = toml.dump(parsed_toml, f)


# processes = set(
#     "julia main_parallel.jl --nproc 10",
#     "julia main_parallel.jl --nproc 10"
# )

# ["python", "-c", "import time; time.sleep(3); print('done')"]

def bag(*args, **kwargs):
    return(asyncio.gather(*args, **kwargs))

def _handle_task_result(task: asyncio.Task) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass  # Task cancellation should not be logged as an error.
    except Exception:  # pylint: disable=broad-except
        logging.exception('Exception raised by task = %r', task)

from rich.console import Console
from time import sleep

class FileSystem:

    def __init__(self) -> None:
        pass

    def exists(self, filename):
        return exists(filename)

"""
This class creates the link between the user and the different executors.
It also contains the main task logic that doesn't have to do with simply running the command, such 
as checking if the task needs to be run, and wether it has run correctly.
"""
class Controller:

    def __init__(self, conf = {}, runner = None) -> None:
        conf = OmegaConf.create(conf)
        self.history = TinyDB('sf.json')
        self.task_queue = queue.Queue()
        self.done   = 0
        self.retry  = 0
        self.failed = 0
        self.last_notify = ""
        self.notifty_hash = conf.get('notify',None)
        self.fs = FileSystem()
        self.force_string = conf.get('force',None)

        self.shut_me_down = False
        self.loop = None

        self.msg_queue = queue.Queue()
        self.logger = logging.getLogger(self.__class__.__name__)

        # we take teh runner if passed (mostly for testing)
        if runner is not None:
            self.executors = { "local": runner }
        elif 'executors' in conf.keys():
            print(conf.executors)
            self.executors = { k:EXECUTORS[k](v) for k,v in conf.executors.items()}
        else:
            self.executors = { "local": CommandRunner({'maxsize':4}) }

    """
    The user tells us that a task should be added
    """
    def add(self, task:Task):

        # check if the task is up to date, 
        # this allows to skip the task right away!
        # if self.check_task_uptodate(task):
        #     self.complete_task(task)
        #     return
        # -> this is done at the time the task is send to the executor

        # otherwise, we add the task to the queue
        self.task_queue.put(task)

    """
    An executor/runner reports a finished task
    """
    def add_completed(self, task:Task) -> None:

        task.set_prop("finish_time", datetime.now().strftime("%d/%m/%Y %H:%M:%S"))

        # retry if output file doesn't exist
        if (len(task.get_outputs())>0) and (not exists(task.get_outputs()[0])):

            if  task.retry>0:
                self.log("[red]output file missing for {}, retrying... [/red]".format(task.uid))  
                task.retry = task.retry - 1
                self.add(task)
            else:
                self.log("[red]output file missing for {}, failing! [/red]".format(task.uid))  
                self.failed = self.failed + 1
                self.complete_task(task)

        # everything normal
        else:
            self.done = self.done +1
            self.complete_task(task)

    """
        Main loop that takes tasks from the global queue and send them to executors.
        This should not send tasks if the executor queue is already full.

        We might want to make this event-driven, ie we check only if:
        - a new task is added
        - a task finished
    """
    def update(self):

        if self.task_queue.empty():
            return

        # we check to see if any exectutor has space in their queues
        for exec in self.executors.values():
            if exec.available_slots()>0:
                # take the next task
                task = self.task_queue.get()

                # check if the task is up to date, 
                # this allows to skip the task right away!
                if self.check_task_uptodate(task):
                    self.complete_task(task)
                    return

                # send it to the executor
                task.set_prop('start_time',datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
                exec.add(task)
                return

    """
    Check if the task is up to date 
    """
    def check_task_uptodate(self, task):

        # add a test to run if matches regexp (this is to force rerun at least once)
        if (self.force_string is not None) and (fnmatch.fnmatch(task.uid, self.force_string)):
            self.log(f"adding forced [red]{task.uid}[/red]")
            self.log(" - cmd: {}".format( " ".join(task.get_command() ) ))
            return(False)

        # checking if the task needs to be redone
        if os.path.exists(task.output_file):
            # console.log("checking {}".format(j.output_file))
            output_time = os.path.getmtime(task.output_file)  

            UP_TO_DATE = True
            for f in task.deps:
                if os.path.getmtime(f) > output_time:    
                    self.log("task [red]{}[/red]: input [green]{}[/green] more recent than output".format(task.uid, f))
                    UP_TO_DATE = False
                    break 
            
            if UP_TO_DATE:
                self.log(f"task [red]{task.uid}[/red] is up to date, skipping it.")
                return(True)
        else:
            self.log(f"adding [red]{task.uid}[/red] output {task.output_file} is missing")

        self.log(f"adding [red]{task.uid}[/red]")
        self.log(" - cmd: {}".format( " ".join(task.get_command() ) ))
        return(False)

    def complete_task(self, task):
        # load the results if any
        # TBD

        self.update_history(task)

        # mark the promise as completed and set it to itself
        assert asyncio.isfuture(task.fut)
        assert not task.fut.done()
        task.set_completed()

    def update_history(self, task):
        # append job to history - replace this with upsert
        # @fixme here sometimes the props don't exist yet, they need to be given a default, or ignored. This happen
        # this happens in particular when skipping the task (then it needs to load first)
        tj = Query()

        newdict = { 
            'deps' : task.deps,
            'output' : task.output_file,
            'cmd': task.get_command()
        }

        # add all available props
        newdict.update(task.props)

        self.history.upsert(newdict, tj.hash == task.hash)

    async def start_loops(self):
        for exec in self.executors.values():
            co_task = asyncio.create_task(exec.loop(self))
            co_task.add_done_callback(_handle_task_result)

        # run my own event loop
        with console.status("Running ...") as status:
            while True:

                # logging messages
                while not self.msg_queue.empty():
                    msg = self.msg_queue.get()
                    console.log(msg)

                # check tasks
                self.update()

                # updating status
                running = 0
                capacity = 0
                for exec in self.executors.values():
                    running += exec.size()
                    capacity += exec.size() + exec.available_slots()

                status.update(f"running [green]queued:{self.task_queue.qsize()} [/green] [yellow]running:{running}/{capacity} [/yellow] [purple]done:{self.done} / failed:{self.failed}[/purple] ...")
                # send updates to sever if any
                self.notify(running, capacity)

                await asyncio.sleep(0.1)

    async def update_loop(self):
        while self.shut_me_down==False:
            self.update()
            await asyncio.sleep(0.1)

    async def shutdown(self):
        self.shut_me_down = True
        await self.loop

    def log(self, str):
        self.msg_queue.put(str)

    """
    notify server
    """
    def notify(self,running, capacity):

        notify_str = "running={}&queued={}&completed={}&failed={}&retry={}".format(
            running, self.task_queue.qsize(), self.done, self.failed, self.retry)

        if notify_str != self.last_notify:
            requests.get('http://scriptflow.lamadon.com/test.php?hash={}&{}'.format(self.notifty_hash, notify_str))
            self.last_notify = notify_str

    """
    called when flow finishes
    """
    def last_word(self):
        console.log("--------------------------------")
        console.log("Thank you for using scriptflow!")
        console.log("You ran {} tasks with {} fails.".format(self.done, self.failed))
        console.log("--------------------------------")

    def set_force_string(self,str):
        self.force_string = str

def init(dict):    
    conf = OmegaConf.create(dict)

    logging.basicConfig(filename='scriptflow.log', level=logging.DEBUG)

    set_main_controller(Controller(conf))
    if conf.debug:
        asyncio.get_event_loop().set_debug(True)